HOURS_IN_DAY: int = 24
SECONDS_IN_MINUTE: int = 60
SECONDS_IN_HOUR: int = SECONDS_IN_MINUTE * 60
SECONDS_IN_DAY: int = SECONDS_IN_HOUR * 24
MS_PER_SEC: int = 1000
MS_IN_DAY: int = SECONDS_IN_DAY * MS_PER_SEC
MS_IN_HOUR: int = 60 * 60 * MS_PER_SEC
MS_IN_MINUTE: int = 60 * MS_PER_SEC
HOURS_IN_WEEK: int = int(24 * 7)
HOURS_IN_YEAR: int = int((SECONDS_IN_DAY * 365.25) // (60 * 60))
